"""
This file describes STATIC layout structure AS COMPILED that comes from the last pass of the `.cub` file compiler.
The general description can be found at .../docs/technote.md
"""

from typing import List, NamedTuple, Dict, Optional, Union
from numbers import Number
from . import veneer, formulae

####################

class Reader:
	"""
	Abstract base class for -- one single (partially-applied) function which
		in particular can be pickled. A bit heavy, but it ain't broke.
	All "sigil" keys get a ComputedReader.
	Rules: If a Frame has an '_' field and a normal-named key, it gets a DefaultReader.
	Other normal-keys get a SimpleReader.
	A menu may not (indeed grammatically cannot) have an '_' field.
	"""
	def __init__(self, key:str):
		assert isinstance(key, str)
		self.key = key

class SimpleReader(Reader): pass
class ComputedReader(Reader): pass
class DefaultReader(Reader): pass

####################

class Hint(NamedTuple):
	boilerplate: formulae.Boilerplate
	priority: int

class Marginalia(NamedTuple):
	""" Corresponds to all the ways you can decorate a shape node. """
	style_index:int = 0
	outline_index:int = 0
	texts:Optional[List[object]] = ()
	hint:Union[int, Hint, None] = None # If this is an integer, it means use the perpendicular header text in that slot.
	height:Optional[Number] = None
	width:Optional[Number] = None

####################

class ShapeDefinition:
	"""
	Abstractly, a static definition for part of the extent (either horizontal
	or vertical) of a class of reports. A recursive data structure and
	potentially a DAG because of the way that partial axis-definitions may be
	linked symbolically in the source language.
	"""

	def __init__(self, margin:Marginalia):
		assert isinstance(margin, Marginalia), type(margin)
		self.margin = margin

	def accumulate_key_space(self, space:set):
		""" Help prepare a set of key-space within the purview of this ShapeDefinition. """
		raise NotImplementedError(type(self))

# There must be at least four kinds of ShapeDefinition: leaves, trees, frames, and menus. Maybe "records" also?

class LeafDefinition(ShapeDefinition):
	
	def accumulate_key_space(self, space: set):
		pass # Nothing to do here.


class CompoundShapeDefinition(ShapeDefinition):
	"""
	Call it implementation inheritance if you must, but it's expedient, factored, and not too weird.
	This deals in the generalities common to Tree, Frame, and Menu.
	Perhaps those differences are one day factored into strategy objects, but for now, it's good enough.
	"""
	def __init__(self, reader:Reader, margin:Marginalia):
		super().__init__(margin)
		self.reader = reader
		self.cursor_key = reader.key
	
	def descend(self, label) -> ShapeDefinition:
		""" This plugs into the planning algorithm. """
		raise NotImplementedError(type(self))
	

class TreeDefinition(CompoundShapeDefinition):
	""" This corresponds nicely to the :tree concept in the language. """
	def __init__(self, reader:Reader, within: ShapeDefinition, margin:Marginalia):
		super().__init__(reader, margin)
		self.within = within
	
	def accumulate_key_space(self, space: set):
		space.add(self.cursor_key)
		self.within.accumulate_key_space(space)
	
	def descend(self, label) -> ShapeDefinition:
		return self.within
	

class FrameDefinition(CompoundShapeDefinition):
	""" This corresponds to a :frame in the language. "Cosmetic" frames may have a reader that returns a constant. """
	def __init__(self, reader:Reader, fields:dict, margin:Marginalia):
		super().__init__(reader, margin)
		self.fields = fields
		self.sequence = list(self.fields.keys())
		pass
	
	def accumulate_key_space(self, space: set):
		# OBSERVATION: Same for FrameDefinition and MenuDefinition (AT THIS TIME).
		space.add(self.cursor_key)
		for within in self.fields.values():
			within.accumulate_key_space(space)
	
	def descend(self, label) -> ShapeDefinition:
		return self.fields[label]
	

class MenuDefinition(CompoundShapeDefinition):
	"""
	A :menu in the language. Has some things in common with both Tree and Frame.
	In principle, you could accomplish a similar concept with "shy fields", perhaps more flexibly.
	But for now, this works.
	"""
	def __init__(self, reader:Reader, fields:dict, margin:Marginalia):
		super().__init__(reader, margin)
		self.fields = fields
	
	def accumulate_key_space(self, space: set):
		# OBSERVATION: Same for FrameDefinition and MenuDefinition (AT THIS TIME).
		space.add(self.cursor_key)
		for within in self.fields.values():
			within.accumulate_key_space(space)
	
	def descend(self, label) -> ShapeDefinition:
		return self.fields[label]

####################

class CanvasDefinition(NamedTuple):
	""" Brings together all the defining characteristics of a single report type. """
	horizontal:ShapeDefinition
	vertical:ShapeDefinition
	background_style:int
	style_rules:List[veneer.Rule[int]]
	formula_rules:List[veneer.Rule[formulae.Formula]]
	merge_specs:List[veneer.Rule[formulae.Formula]]

class OutlineData(NamedTuple):
	level: int
	hidden: bool
	collapsed: bool

class CubModule(NamedTuple):
	"""
	The uppermost final compiled object: what you expect to get back from "compiling" a `.cub` file.
	You SHOULD be able to pickle one of these objects to avoid repeat-compilation.
	"""
	canvases: Dict[str, CanvasDefinition]
	styles: List[dict]
	outlines: List[OutlineData]




from typing import List, Tuple
from boozetools.support import runtime as brt
from boozetools.support.interfaces import Scanner
from . import AST, formulae, utility

TABLES = utility.tables(__file__, 'core.md')

class CoreDriver(brt.TypicalApplication):
	VALID_KEYWORDS = {'AXIS', 'CANVAS', 'FRAME', 'GAP', 'HEAD', 'LEAF', 'MENU', 'MERGE', 'OF', 'USE', 'TREE', 'STYLE'}
	
	def default_scan_action(self, message, scanner, param):
		# Just in case I forgot something:
		raise TypeError("scan_%s needs to be defined now." % message)
	
	def scan_ignore(self, yy:Scanner):
		assert '\n' not in yy.matched_text()
	
	def scan_keyword(self, yy:Scanner):
		word = yy.matched_text()[1:].upper()
		if word not in self.VALID_KEYWORDS: word='$bogus$'
		yy.token(word, None)
	
	def scan_enter(self, yy:Scanner, dst):
		yy.push(dst)
		yy.token('BEGIN_'+dst, None)
	
	def scan_leave(self, yy:Scanner, src):
		yy.pop()
		yy.token('END_'+src, None)
	
	def scan_punctuation(self, yy:Scanner):
		it = yy.matched_text()
		yy.token(it, it)
	
	def scan_embedded_newline(self, yy:Scanner):
		yy.token('TEXT', AST.Constant('\n', yy.current_span(), 'TEXT'))
		
	def scan_letter_escape(self, yy:Scanner):
		yy.token('TEXT', chr(7+'abtnvfr'.index(yy.matched_text())))
	
	def scan_token(self, yy:Scanner, kind):
		# Concrete tokens that won't contribute to the AST
		yy.token(kind, yy.matched_text())
	
	def scan_name(self, yy:Scanner, kind):
		# sort of like an identifier: We need to capture the span in case of semantic error.
		yy.token(kind, AST.Name(yy.matched_text(), yy.current_span()))
	
	def scan_string(self, yy:Scanner, kind):
		# Pick up the token as a constant of string type, not an identifier.
		# Like a bareword in ancient perl?
		yy.token(kind, AST.Constant(yy.matched_text(), yy.current_span(), kind))
	
	def scan_escape_literal(self, yy:Scanner):
		yy.token('TEXT', AST.Constant(yy.matched_text()[1:], yy.current_span(), 'TEXT'))
	
	def scan_sigil(self, yy:Scanner, kind:str):
		# Like scan_string but removes the first char from the semantic value.
		name = AST.Name(yy.matched_text()[1:], yy.current_span())
		yy.token(kind, AST.Sigil(name, kind))
	
	def scan_integer(self, yy:Scanner):
		yy.token('INTEGER', AST.Constant(int(yy.matched_text()), yy.current_span(), 'INT'))
	
	def scan_hex_integer(self, yy:Scanner):
		yy.token('INTEGER', AST.Constant(int(yy.matched_text()[1:], 16), yy.current_span(), 'HEX'))
	
	def scan_decimal(self, yy:Scanner):
		yy.token('DECIMAL', AST.Constant(float(yy.matched_text()), yy.current_span(), 'DEC'))

	def scan_delimited(self, yy:Scanner, kind):
		yy.token(kind, AST.Constant(yy.matched_text()[1:-1], yy.current_span(), 'STRING'))

	def __init__(self):
		self.errors=0
		super().__init__(TABLES)
	
	def unexpected_token(self, kind, semantic, pds):
		self.errors += 1
		print("in scan state:", self.yy.current_condition())
		breadcrumbs = TABLES['parser']['breadcrumbs']
		t = TABLES['parser']['terminals']
		nt = TABLES['parser']['nonterminals']
		def bc(q):
			x = breadcrumbs[q]
			if x < len(t): return t[x]
			else: return nt[x-len(t)]
		stack_symbols = list(map(bc, list(pds.path_from_root())[1:]))
		print("Parse Stack:", *stack_symbols)
		super().unexpected_token(kind, semantic, pds)
	
	def parse_none(self): return None
	def parse_true(self, *args): return True
	def parse_false(self, *args): return False
	def parse_empty(self): return []
	def parse_singleton(self, item): return [item]
	def parse_append(self, them, item):
		them.append(item)
		return them
	
	def parse_define_style(self, name:AST.Name, elts:list):
		assert isinstance(name, AST.Name), type(name)
		return AST.StyleDef(name, elts)
	
	def parse_marginalia(self, texts, hint, appearance):
		return AST.Marginalia(texts, hint, appearance)
	
	def parse_field(self, name, shape):
		assert isinstance(name, AST.Name), type(name)
		assert isinstance(shape, (AST.Marginalia, AST.Frame, AST.Menu, AST.Tree, AST.LinkRef)), type(shape)
		return AST.Field(name, shape)
	
	def parse_frame(self, margin, key, fields): return AST.Frame(margin, key, fields)
	def parse_menu(self, margin, key, fields): return AST.Menu(margin, key, fields)
	def parse_tree(self, margin, key, within): return AST.Tree(margin, key, within)
	
	def parse_assignment(self, name, value):
		assert isinstance(name, AST.Name)
		if isinstance(value, AST.Name): value = AST.Constant(*value, 'STRING')
		assert isinstance(value, AST.Constant)
		return AST.Assign(name, value)
	
	def parse_define_canvas(self, name:AST.Name, across:AST.Name, down:AST.Name, style_points:list, items:list):
		return AST.Canvas(name, across, down, style_points, items)
	
	def parse_label_interpolated(self, items) -> formulae.Boilerplate:
		if len(items) == 1 and isinstance(items[0], formulae.RawOrdinal):
			return formulae.RawCell(items[0].axis)
		else:
			return formulae.Label(items)
	
	def parse_label_constant(self, text:AST.Constant) -> formulae.Label:
		assert text.kind == 'STRING'
		assert isinstance(text.value, str)
		return formulae.Label([formulae.LiteralText(text.value)])
	
	def parse_literal_text(self, text:AST.Constant):
		assert text.kind == 'TEXT'
		return formulae.LiteralText(text.value)
	def parse_tpl_plaintext(self, name:AST.Name): return formulae.PlainOrdinal(name.text)
	def parse_tpl_raw(self, name:AST.Name): return formulae.RawOrdinal(name.text)
	def parse_select_set(self, fields:List[AST.Name]) -> formulae.Predicate:
		if len(fields) == 1: return formulae.IsEqual(fields[0].text)
		else: return formulae.IsInSet(frozenset(f.text for f in fields))
	def parse_select_not_set(self, fields:List[AST.Name]) -> formulae.Predicate:
		return formulae.IsNotInSet(frozenset(f.text for f in fields))
	def parse_select_computed(self, computed:AST.Sigil) -> formulae.ComputedPredicate :
		assert isinstance(computed, AST.Sigil)
		assert computed.kind == 'COMPUTED', computed.kind
		return formulae.ComputedPredicate(computed.name.text)
	def parse_select_each(self) -> formulae.IsDefined:
		return formulae.IsDefined()
	def parse_criterion(self, field_name:AST.Name, predicate:formulae.Predicate):
		# TODO: You can make the argument that a name on the left-hand side of
		#  a criterion ought to appear in any canvas that uses it. And sure,
		#  that would be nice to validate. But not right this minute.
		return field_name.text, predicate
	def parse_selector(self, criteria:List[Tuple[str, formulae.Predicate]]):
		cd = {}
		for axis, predicate in criteria:
			if axis not in cd: cd[axis] = predicate
			elif isinstance(cd[axis], list): cd[axis].append(predicate)
			else: cd[axis] = [cd[axis], predicate]
		return formulae.Selection(cd)
	def parse_formula(self, fragments): return formulae.Formula(fragments)
	def parse_gap_hint(self): return AST.GAP_HINT
	def parse_linkref(self, marginalia:AST.Marginalia, name:AST.Name):
		return AST.LinkRef(marginalia, name)



"""
Library of AST nodes built by the grammar compiler and walked by the transducer.
This file is an organizational tool as much as anything.
One thing: For a post-parse pass to show errors in context, it's necessary to
pass location information around in all the right places.
"""

from typing import NamedTuple, List, Union, Tuple
from . import formulae

class Name(NamedTuple):
	text: str
	span: Tuple[int, int]

class Sigil(NamedTuple):
	name: Name
	kind: str

class Constant(NamedTuple):
	value: object
	span: Tuple[int, int]
	kind: str

class Assign(NamedTuple):
	# word is either a formatting or outlining property
	word: Name
	const: Constant

class StyleDef(NamedTuple):
	name: Name
	elts: list

GAP_HINT = object()

class Marginalia(NamedTuple):
	texts: list
	hint: object # If an integer, then a head-hint. If GAP_HINT, a gap. If a tuple, a function or formula and priority.
	appearance: List[Union[Assign, Name]] # Words shall mean style references.

class Field(NamedTuple):
	name:Name
	shape:object # Marginalia -> Leaf; Name -> named shape; Frame/Menu/Tree: define as such.

class Frame(NamedTuple):
	margin:Marginalia
	key:Union[Name, Sigil, None]
	fields:list
	
class Menu(NamedTuple):
	margin:Marginalia
	key:Union[Name, Sigil, None]
	fields:list
	
class Tree(NamedTuple):
	margin:Marginalia
	key:Union[Name, Sigil, None]
	within:object

class LinkRef(NamedTuple):
	margin: Marginalia
	name: Name
	
class Canvas(NamedTuple):
	name:Name
	across:Name
	down:Name
	style_points:list
	patches:list




"""
Here shall be defined those things necessary to integrate the presentation system
with whatever run-time plug-in computing power.
"""

from typing import Mapping

class DataStreamError(KeyError):
	pass

class AbsentKeyError(DataStreamError):
	""" args[0] is the missing Identifier. args[1] is the offending point. """

class InvalidOrdinalError(DataStreamError):
	""" args[0] is the axis Identifier, or None. args[1] is the offending ordinal. """



class Environment:
	"""
	In particular, this is the bit that gets passed around as a sort of "outside-the-global" scope.
	"""
	
	def test_predicate(self, predicate_name, ordinal) -> bool:
		method = getattr(self, 'is_'+predicate_name)
		return method(ordinal)
	
	def collation(self, dimension_name):
		""" If desired, return a sort-key function for whatever dimension. """
		return getattr(self, 'collate_'+dimension_name, None)
	
	def read_magic(self, key, point:Mapping):
		method = getattr(self, 'magic_'+key)
		return method(point)

	def plain_text(self, key, value):
		""" This is the part where you'd create special "friendly-text" methods for different things. """
		return str(value)
	


"""
This file tells you what your options and flexibilities are in formatting an XLSX spreadsheet.
It has sections on cell formatting and outline formatting.
"""

import re

class Kind:
	""" This is sort of like a type. But it's allowed to be more specific. """
	def __init__(self, description:str, test):
		self.description = description
		self.test = test

class RangeKind(Kind):
	def __init__(self, lower, upper):
		super().__init__('an integer from %d to %d, inclusive', lambda x:isinstance(x,int) and lower <= x <= upper)

kind_string = Kind('a string', lambda x:isinstance(x,str))
kind_number = Kind('a number', lambda x:isinstance(x,(int, float)))
kind_integer = Kind('a number', lambda x:isinstance(x,int))
kind_boolean = Kind('a true/false flag', lambda x:isinstance(x,bool))
kind_border = RangeKind(0, 13)

COLOR_NAMES = frozenset('black blue brown cyan gray green lime magenta navy orange pink purple red silver white yellow'.split())
kind_color = Kind('a valid color name or #hex code', lambda x:isinstance(x, str) and (x in COLOR_NAMES or re.fullmatch(r'#[0-9A-Fa-f]{6}', x)))

FORMAT_PROPERTIES = {
	'font_name': kind_string,
	'font_size': kind_integer,
	'font_color': kind_string,
	'bold': kind_boolean,
	'italic': kind_boolean,
	'underline': Kind('one of 1=single, 2=double, 33=single-accounting, 34=double-accounting', {None,1,2,33,34}.__contains__),
	'font_strikeout': kind_boolean,
	'font_script': Kind('one of 1=Superscript, 2=Subscript', {1,2}.__contains__), #  -- but these are unlikely.
	'num_format': Kind("a valid format string or numeric code", lambda x:(isinstance(x,int) and 0<=x<=49) or (isinstance(x,str))),
	'locked': kind_boolean,
	'hidden': kind_boolean,
	'align': kind_string,
	'valign': kind_string,
	'rotation': Kind('from -90 to +90 or exactly 270', lambda x: isinstance(x, (int, float)) and ((-90 <= x <= 90) or (x == 270))),
	'text_wrap': kind_boolean,
	'reading_order': Kind('one of 1=left-to-right, like English. 2=right-to-left, like Arabic', {1,2}.__contains__), # .
	'text_justlast': kind_boolean,
	# 'center_across': # use align=center_across instead.
	'indent': kind_integer,
	'shrink': kind_boolean,
	'pattern': RangeKind(0,18), # 0-18, with 1 being a solid fill of the background color.
	'bg_color': kind_color,
	'fg_color': kind_color,
	'border': kind_border,
	'bottom': kind_border,
	'top': kind_border,
	'left': kind_border,
	'right': kind_border,
	'diag_border': kind_border,
	'diag_type': RangeKind(0,3),
	'diag_color': kind_color,
	'border_color': kind_color,
	'bottom_color': kind_color,
	'top_color': kind_color,
	'left_color': kind_color,
	'right_color': kind_color,
}

OUTLINE_PROPERTIES = {
	'height': kind_number,
	'width': kind_number,
	'level': RangeKind(0,7),
	'hidden': kind_boolean,
	'collapsed': kind_boolean,
}

SPECIAL_CASE = {
	'border': ['top', 'left', 'bottom', 'right'],
	'border_color': ['top_color', 'left_color', 'bottom_color', 'right_color'],
}




"""
This module focuses on the transformation from AST to static structure.
There's not really too much magic here, but it's good to have the separation.
It's also completely unusable at the moment; just a parking lot for excised code.
But things will get better...
"""

import collections
from typing import List, Union, Mapping, MutableMapping
from boozetools.support import foundation, failureprone
from . import AST, static, formulae, xl_schema, veneer


class SemanticError(Exception):
	"""
	Catch-all for things that should not be.
	"""

class UndefinedNameError(KeyError): pass
class RedefinedNameError(KeyError): pass
class BadAttributeValue(ValueError): pass
class NoSuchAttrbute(KeyError): pass

BLANK_STYLE = collections.ChainMap({
	'_texts':(), '_hint':None,
	'level':0, 'hidden':False, 'collapsed':False,
	'height':None, 'width':None,
})

VOCABULARY = {
	**xl_schema.FORMAT_PROPERTIES,
	**xl_schema.OUTLINE_PROPERTIES,
	**{k: xl_schema.FORMAT_PROPERTIES[v[0]] for k, v in xl_schema.SPECIAL_CASE.items()},
}


class SymbolTable:
	""" I've a sense I'll want a "smart" symbol table object: a bit more than a dictionary. """
	def __init__(self):
		self.__entries = {} # For now entries are just a <Name, object> pair. Keys are strings.
	
	def __contains__(self, item):
		if isinstance(item, str): return item in self.__entries
		if isinstance(item, AST.Name): return item.text in self.__entries
		raise TypeError(item)
	
	def let(self, name:AST.Name, item):
		""" Enforce single-assignment... """
		assert isinstance(name, AST.Name), type(name)
		key = name.text
		if key in self.__entries: raise RedefinedNameError(name, self.__entries[key][0])
		else: self.__entries[key] = (name, item)
	
	def get(self, name:AST.Name):
		try: return self.__entries[name.text][1]
		except KeyError: raise UndefinedNameError(name) from None
	
	def get_declaration(self, key:str) -> AST.Name:
		""" Go find where was the name that declared the symbol. """
		return self.__entries[key][0]
	
	def as_dict(self) -> dict:
		""" Just string keys and semantic items; no location-tracking fluff. """
		return {key: entry[1] for key, entry in self.__entries.items()}
	

class Transducer(foundation.Visitor):
	"""
	A form of interpreter: not from source lines of code, but from a list of
	AST nodes for top-level declarations. It can also be considered a kind of
	tree-transducer, if you squint.
	
	Someone will object that this should not be written as an object (I pun)
	but the alternative (a nest of procedures) gets ugly fast.
	"""
	def __init__(self, source:failureprone.SourceText):
		self.source = source
		
		self.named_styles = SymbolTable()
		self.named_fields = SymbolTable()
		self.translated_shapes = SymbolTable()
		self.named_canvases = SymbolTable()
		
		self.numbered_styles = foundation.EquivalenceClassifier()
		self.numbered_outlines = foundation.EquivalenceClassifier()
	
	def interpret(self, declarations:List[Union[AST.StyleDef, AST.Field, AST.Canvas]]) -> static.CubModule:
		for d in declarations: self.visit(d)
		return static.CubModule(
			self.named_canvases.as_dict(),
			self.numbered_styles.exemplars,
			self.numbered_outlines.exemplars,
		)
	
	def visit_Field(self, field:AST.Field):
		self.named_fields.let(field.name, field.shape)

	def visit_Canvas(self, canvas:AST.Canvas):
		""" Build a static.CanvasDefinition and add it to self.named_canvases """
		def mk_style_rule(sel, elts):
			style_rules.append(veneer.Rule(sel, style_index(elts)))
		
		def style_index(elts):
			env = {}
			self.build_style(elts, env)
			return self.make_numbered_style(env)

		style_rules = []
		formula_rules = []
		merge_rules = []
		
		background_style = style_index(canvas.style_points)
		for is_merge, selector, content, style in canvas.patches:
			assert isinstance(selector, formulae.Selection)
			if style: mk_style_rule(selector, style)
			if is_merge: merge_rules.append(veneer.Rule(selector, content))
			elif content: formula_rules.append(veneer.Rule(selector, content))
		self.named_canvases.let(canvas.name, static.CanvasDefinition(
			horizontal=self.get_translated(canvas.across),
			vertical=self.get_translated(canvas.down),
			background_style=background_style,
			style_rules=style_rules,
			formula_rules=formula_rules,
			merge_specs=merge_rules,
		))
		pass
	
	def get_translated(self, name:AST.Name) -> static.ShapeDefinition:
		if name not in self.translated_shapes:
			shape = self.named_fields.get(name)
			item = FieldBuilder(self, BLANK_STYLE, name).visit(shape)
			self.translated_shapes.let(name, item)
		return self.translated_shapes.get(name)
	
	def visit_StyleDef(self, styledef:AST.StyleDef):
		"""
		Build up a named-style as a root-level object and store it in the style namespace.
		At least, that's the happy path. Various things can go wrong. What gets reported
		first is sort of a matter of happenstance: what mistake do we stumble over first?
		At any rate, report definitions should not ever get huge, so reporting an error
		and quitting is probably an OK strategy for now.
		"""
		env = {}
		self.build_style(styledef.elts, env)
		try: self.named_styles.let(styledef.name, env)
		except RedefinedNameError as e:
			self.source.complain(*e.args[0].span, message="Redefined style name")
			self.source.complain(*e.args[1].span, message="First defined here")
			raise
	
	def build_style(self, elts:List, env:MutableMapping):
		"""
		elts are -- style names, activations, deactivations, or attribute assignments.
		This works by mutating an environment you pass in, so the same code therefore works
		regardless of use in creating named-styles or actually-used styles.
		"""
		def try_assign(name:AST.Name, value, value_span):
			"""
			The key must be a properly recognized style property as defined in module `xl_schema`.
			There are two special cases (`border` and `border_color`) which behave as if corresponding
			top, left, bottom, and right properties are all set at once.
			The value must be consistent with the expectations for the property the key names.
			The spans give locations in a source text to complain about if something goes wrong.
			"""
			key = name.text
			if key not in VOCABULARY:
				self.source.complain(*name.span, message="There's no such attribute as '%s'." % key)
				raise NoSuchAttrbute(key)
			kind = VOCABULARY[key]
			if not kind.test(value):
				self.source.complain(*value_span, message="Value for %s must be %s" % (key, kind.description))
				raise BadAttributeValue(value)
			if key in xl_schema.SPECIAL_CASE:
				for k in xl_schema.SPECIAL_CASE[key]: env[k] = value
			else: env[key] = value
		
		for item in elts:
			if isinstance(item, AST.Assign):
				try_assign(item.word, item.const.value, item.const.span)
			elif isinstance(item, AST.Sigil):
				if item.kind == 'ACTIVATE': try_assign(item.name, True, item.name.span)
				elif item.kind == 'DEACTIVATE': try_assign(item.name, False, item.name.span)
				elif item.kind == 'STYLE_NAME': env.update(self.named_styles.get(item.name))
			else: assert False, item
	
	def make_numbered_style(self, env:Mapping) -> int:
		digest = sorted((k, v) for k, v in env.items() if k in xl_schema.FORMAT_PROPERTIES)
		return self.numbered_styles.classify(tuple(digest))
	
	def make_numbered_outline(self, env:Mapping) -> int:
		digest = (env['level'], env['hidden'], env['collapsed'])
		return self.numbered_outlines.classify(digest)

class FieldBuilder(foundation.Visitor):
	"""
	I have this idea that the answer to translating fields is properly recursive only on
	a slightly reduced form of the original problem.
	"""
	def __init__(self, module_builder:Transducer, style_context:collections.ChainMap, name:AST.Name):
		self.mb = module_builder
		self.sc = style_context.new_child()
		self.name = name
	
	def subordinate(self, name:AST.Name) -> "FieldBuilder":
		return FieldBuilder(self.mb, self.sc, name)
	
	def interpret_margin_notes(self, notes:AST.Marginalia) -> static.Marginalia:
		"""
		Two phases: Update the fields of the style context (self.sc) according to the notes,
		and then capture the result in the appointed structure.
		"""
		# FIXME: This neglects to transduce the texts and hints.
		if notes.texts is not None:
			self.sc['_texts'] = notes.texts
		if notes.hint is not None:
			self.sc['_hint'] = self.interpret_hint(notes.hint)
		self.mb.build_style(notes.appearance, self.sc)
		return static.Marginalia(
			style_index=self.mb.make_numbered_style(self.sc),
			outline_index=self.mb.make_numbered_outline(self.sc),
			texts=self.sc['_texts'],
			hint=self.sc['_hint'],
			height=self.sc['height'],
			width=self.sc['width'],
		)
	
	def interpret_hint(self, hint:object):
		if hint is AST.GAP_HINT: return static.Hint(formulae.THE_NOTHING, 999)
		if isinstance(hint, AST.Constant) and hint.kind == 'INT':
			# It's a reference to a perpendicular text. It has precedence above marginal formulas.
			assert isinstance(hint.value, int), hint
			assert hint.value > 0, hint.value
			return hint.value - 1
		boilerplate, prio_spec = hint
		if prio_spec is None: priority = 0
		elif isinstance(prio_spec, AST.Constant):
			assert prio_spec.kind == 'INT'
			priority = prio_spec.value
		else: assert False, type(prio_spec)
		assert isinstance(priority, int)
		return static.Hint(boilerplate, priority)
	
	def visit_Marginalia(self, notes:AST.Marginalia) -> static.LeafDefinition:
		return static.LeafDefinition(self.interpret_margin_notes(notes))
	
	def visit_LinkRef(self, linkref:AST.LinkRef) -> static.ShapeDefinition:
		sub = self.subordinate(linkref.name)
		sub.interpret_margin_notes(linkref.margin)
		return sub.visit(self.mb.named_fields.get(linkref.name))
	
	def visit_Tree(self, tree:AST.Tree) -> static.TreeDefinition:
		margin = self.interpret_margin_notes(tree.margin)
		key = tree.key or self.name
		if isinstance(key, AST.Name):
			reader = static.SimpleReader(key.text)
		elif isinstance(key, AST.Sigil):
			assert key.kind == 'COMPUTED', key.kind
			reader = static.ComputedReader(key.name.text)
		else:
			assert False, type(key)
		within = self.visit(tree.within)
		return static.TreeDefinition(reader, within, margin)
	
	def visit_Frame(self, frame:AST.Frame) -> static.FrameDefinition:
		# Expanding on this first because it's called first.
		# The margin notes are the easy bit:
		margin = self.interpret_margin_notes(frame.margin)
		children = self.__children(frame.fields)
		
		# There are three possible key types: name, sigil, and None.
		key = frame.key or self.name
		if isinstance(key, AST.Name):
			axis = key.text
			if '_' in children: reader = static.DefaultReader(axis)
			else: reader = static.SimpleReader(axis)
			# However, if there's a field called '_', then we want a default-reader instead?
		elif isinstance(key, AST.Sigil):
			assert key.kind == 'COMPUTED', key.kind
			try: symbol = children.get_declaration('_')
			except KeyError: pass
			else:
				self.mb.source.complain(*key.name.span, message="This tells me the environment will supply a field name...")
				self.mb.source.complain(*symbol.span, message="Therefore, a 'default' field makes no sense.")
				raise SemanticError()
			reader = static.ComputedReader(key.name.text)
		else:
			assert False, type(key)
		
		return static.FrameDefinition(reader, children.as_dict(), margin)
	
	def __children(self, fields:list) -> SymbolTable:
		# Recursion on the children should be a useful trick...
		children = SymbolTable()
		for symbol, definition in fields:
			children.let(symbol, self.subordinate(symbol).visit(definition))
		return children
	
	def visit_Menu(self, menu:AST.Menu) -> static.MenuDefinition:
		# Expanding on this first because it's called first.
		# The margin notes are the easy bit:
		margin = self.interpret_margin_notes(menu.margin)
		children = self.__children(menu.fields)
		assert '_' not in children, "This is grammatically impossible!"
		
		# There are three possible key types: name, sigil, and None.
		key = menu.key or self.name
		if isinstance(key, AST.Name):
			reader = static.SimpleReader(key.text)
		elif isinstance(key, AST.Sigil):
			assert key.kind == 'COMPUTED', key.kind
			reader = static.ComputedReader(key.name.text)
		else:
			assert False, type(key)
		
		return static.MenuDefinition(reader, children.as_dict(), margin)




"""
This file describes the DYNAMIC structure of a fully instantiated object DEFINED in a `.cub` file.
There is a bit of an odd dependency loop here. To resolve it, I think it's necessary to factor out
the common data structure which is a "layout tree".

The general description can be found at .../docs/technote.md
"""

import collections
from typing import Optional, Dict, Callable, List
from boozetools.support import foundation
from . import static, formulae, runtime, veneer, utility


class Node:
	""" This is a dynamic layout node, but it's much more a common language for the static and dynamic parts. """

class LeafNode(Node):
	""" Seems a half-decent idea to distinguish... """
	__slots__ = ["begin", "margin", "style_class", "formula_class"]
	def __init__(self, margin:static.Marginalia):
		self.margin = margin
	def end(self): return self.begin
	def after(self): return self.begin+1

class InternalNode(Node):
	""" This can STILL be empty... """
	__slots__ = ["begin", "margin", "style_class", "formula_class", "children", "size"]
	def __init__(self, margin:static.Marginalia):
		self.margin = margin
		self.children = {}
	def end(self): return self.begin + self.size - 1
	def after(self): return self.begin + self.size


class Canvas:
	"""
	This is the actual object that collects actual data for actual plotting into an actual spreadsheet somewhere.
	It naturally refers back to its definition structure, but also to relevant per-instance data.
	
	Stylistic note: It would be entirely possible to host all functionality within the structure
	and require client code to pass in both but then client code might accidentally get it wrong...
	It's better this way I think. At least for the overall canvas object.
	"""
	def __init__(self, cub_module:static.CubModule, identifier:str, environment:runtime.Environment):
		self.cub_module = cub_module # This turns out to get consulted...
		self.definition = cub_module.canvases[identifier]
		self.environment = environment
		self.cell_data = collections.defaultdict(int)
		self.across = Direction(self.definition.horizontal, environment)
		self.down = Direction(self.definition.vertical, environment)
		self.space = self.across.space | self.down.space
		intersection = self.across.space & self.down.space
		assert not intersection, intersection
		
	# A few routines for plugging data into a grid:
	
	def poke(self, point, value):
		self.cell_data[self.key_pair(point)] = value
	
	def incr(self, point, value):
		self.cell_data[self.key_pair(point)] += value
	
	def decr(self, point, value):
		self.cell_data[self.key_pair(point)] -= value
	
	def key_pair(self, point):
		fkn = FindKeyNode(point, self.environment)
		return fkn.visit(self.across), fkn.visit(self.down)
	
	# It's sometimes necessary to remove rows and/or columns that are, for instance, all zero or nearly so.
	# The relevant
	
	# Since all the cosmetic surgery is performed in the language, it's therefore all represented in
	# the .definition object, and thus we can proceed on to plotting.
	
	def plot(self, workbook, sheet, top_row_index:int, left_column_index:int, blank=None):
		""" Argument order (row/column) is here consistent with xlsxwriter. """
		
		background_format = self.cub_module.styles[self.definition.background_style]
		
		def find_format():
			"""
			For styling, the concept is simple enough: You take the margin styles as a background and then overlay
			that with any extra bits that are specified in the canvas definition as style rules.
			"""
			col_style = col_margin.style_index
			row_style = row_margin.style_index
			col_cls = col_node.style_class
			row_cls = row_node.style_class
			fmt_key = col_style, row_style, col_cls, row_cls
			try: return style_cache[fmt_key]
			except:
				styles = self.cub_module.styles
				rules = self.definition.style_rules
				bits = dict(background_format)
				bits.update(styles[row_style])
				bits.update(styles[col_style])
				for i in skin.select(col_cls, row_cls):
					bits.update(styles[rules[i].payload])
				it = style_cache[fmt_key] = workbook.add_format(bits)
				return it
			pass
		
		def template(index, yon:static.Marginalia):
			if isinstance(index, int):
				them = yon.texts
				if index < len(them): return them[index]
				else: return formulae.THE_NOTHING
		
		def compete(a:Optional[static.Hint], b:Optional[static.Hint]):
			if a is None: return b
			if b is None: return a
			if b.priority > a.priority: return b
			return a
		
		def find_formula():
			"""
			Determining which hint applies is a bit more of a trick.
			First, if there's a patch defined which applies, then it takes priority.
			Otherwise, if a margin's "hint" field contains an integer, that's a
			reference to the OTHER axis's corresponding list of margin templates.
			Next, one margin.hint may supply a specific hint to use (with priority)
			"""
			cf, rf = col_margin.hint, row_margin.hint
			if 'gap' in (cf, rf): return None
			formula = check_patch()
			if formula is None: formula = template(cf, row_margin)
			if formula is None: formula = template(rf, col_margin)
			if formula is None: formula = compete(cf, rf)
			if formula is None: return self.cell_data.get((col_node, row_node), blank)
			return FormulaInterpreter(cursor, self).visit(formula)
		
		def check_patch():
			patch_key = col_node.formula_class, row_node.formula_class
			try: return formula_cache[patch_key]
			except KeyError:
				candidates = patch.select(*patch_key)
				if candidates:
					winning_patch_index = candidates[-1]
					winning_rule = self.definition.formula_rules[winning_patch_index]
					winning_formula = winning_rule.payload
				else:
					winning_formula = None
				it = formula_cache[patch_key] = winning_formula
				return it
		
		def ops(index):
			level, hidden, collapsed = self.cub_module.outlines[index]
			return {'level':level, 'hidden':hidden, 'collapsed':collapsed}
		
		style_cache = {}
		formula_cache = {}
		skin = veneer.CrossClassifier(self.definition.style_rules, self.across.space, self.down.space)
		patch = veneer.CrossClassifier(self.definition.formula_rules, self.across.space, self.down.space)
		self.across.plan(Cartographer(left_column_index, skin.across, patch.across))
		self.down.plan(Cartographer(top_row_index, skin.down, patch.down))
		cursor = {}
		tour = LeafTour(cursor)
		# Set all the widths etc.
		for col_node in tour.visit(self.across):
			assert isinstance(col_node, LeafNode)
			col_margin = col_node.margin
			sheet.set_column(col_node.begin, col_node.begin, col_margin.width, options=ops(col_margin.outline_index))
		
		# Set all the heights etc. and plot all the data.
		for row_node in tour.visit(self.down):
			assert isinstance(row_node, LeafNode)
			row_margin = row_node.margin
			sheet.set_row(row_node.begin, row_margin.height, options=ops(row_margin.outline_index))
			
			for col_node in tour.visit(self.across):
				assert isinstance(col_node, LeafNode)
				col_margin = col_node.margin
				sheet.write(row_node.begin, col_node.begin, find_formula(), find_format())
			pass
		
		# Plot all merge cells rules. This is done literally in order of merge rules.
		# Formats on the merge cells are computed the same way as those on regular cells.
		for spec in self.definition.merge_specs:
			across = spec.selection.projection(self.across.space)
			for row_node in self.down.tour_merge(cursor, spec.selection.projection(self.down.space)):
				row_margin = row_node.margin
				top,bottom = row_node.begin, row_node.end()
				for col_node in self.across.tour_merge(cursor, across):
					col_margin = col_node.margin
					left,right = col_node.begin, col_node.end()
					item = FormulaInterpreter(cursor, self).visit(spec.payload)
					if top==bottom and left==right: sheet.write(top, left, item, find_format())
					else: sheet.merge_range(top, left, bottom, right, item, find_format())
		pass
	
	def data_range(self, cursor, selection:formulae.Selection):
		"""
		All the DATA cells where all criteria are met, as a list of ranges or cells (or just a zero)
		"""
		assert selection.criteria.keys() <= self.space, '%r not found among %r'%(selection.criteria.keys() - self.space, self.space)
		columns = self.across.data_index(cursor, selection)
		rows = self.down.data_index(cursor, selection)
		if rows and columns:
			return [utility.make_range(c, r) for c in columns for r in rows]
		else:
			return ["0"]

class FormulaInterpreter(foundation.Visitor):
	"""
	Not sure if this needs to be its own class or methods on Canvas,
	but this works OK for now.
	"""
	def __init__(self, cursor:dict, canvas:Canvas):
		self.cursor = cursor
		self.canvas = canvas
		self.env = canvas.environment
	
	def visit_BlankCell(self, _:formulae.BlankCell): return None
	
	def visit_Label(self, label:formulae.Label):
		return ''.join(str(self.visit(e)) for e in label.bits)
	
	def visit_LiteralText(self, literal:formulae.LiteralText):
		return literal.text
	
	def visit_PlainOrdinal(self, sub:formulae.PlainOrdinal):
		key = sub.axis
		value = self.cursor[key]
		return self.env.plain_text(key, value)
	
	def visit_Hint(self, hint:static.Hint):
		return self.visit(hint.boilerplate)
	
	def visit_Formula(self, formula:formulae.Formula):
		return '='+''.join(str(self.visit(e)) for e in formula.bits)
	
	def visit_Selection(self, selection:formulae.Selection):
		return ','.join(self.canvas.data_range(self.cursor, selection))

class Direction:
	"""
	Turns out there's a whole bunch of stuff where you have to keep the right dynamic tree with the right
	static definition. There are two of each for any given grid. Thus, rather than have a whole mess of
	correspondences where I can accidentally get it wrong, I'll embody that correspondence as a single
	unit of meaning in the form of this class.
	"""
	def __init__(self, shape:static.ShapeDefinition, env:runtime.Environment):
		self.shape = shape
		self.env = env
		self.tree = node_factory.visit(self.shape)
		self.space = set()
		shape.accumulate_key_space(self.space)

	def plan(self, cartographer:"Cartographer"):
		state = veneer.PlanState({}, frozenset(), frozenset(), self.env)
		cartographer.visit(self.shape, self.tree, state)
	
	def data_index(self, cursor, selection:formulae.Selection):
		fd = FindData(cursor, selection.projection(self.space))
		try: fd.visit(self.shape, self.tree, len(fd.criteria))
		except runtime.AbsentKeyError as ake:
			print(selection)
			raise
		return utility.collapse_runs(sorted(fd.found))
	
	def tour_merge(self, cursor, selection:formulae.Selection):
		return InternalTour(cursor, selection).visit(self.shape, self.tree, len(selection.criteria))

class FindKeyNode(foundation.Visitor):
	""" Go find the appropriate sub-node for a given point, principally for entering magnitude/attribute data. """
	
	def __init__(self, point: dict, env:runtime.Environment):
		self.point = point
		self.env = env
	
	def visit_Direction(self, direction: Direction):
		return self.visit(direction.shape, direction.tree)
	
	def visit_LeafDefinition(self, shape:static.LeafDefinition, node:LeafNode) -> LeafNode:
		return node
	
	def visit_TreeDefinition(self, shape:static.TreeDefinition, node:InternalNode) -> LeafNode:
		ordinal = self.visit(shape.reader)
		try: branch = node.children[ordinal]
		except KeyError: branch = node.children[ordinal] = node_factory.visit(shape.within)
		return self.visit(shape.within, branch)
	
	def visit_FrameDefinition(self, shape:static.FrameDefinition, node:InternalNode) -> LeafNode:
		ordinal = self.visit(shape.reader)
		try: branch = node.children[ordinal]
		except KeyError: raise runtime.InvalidOrdinalError(shape.cursor_key, ordinal)
		else: return self.visit(shape.fields[ordinal], branch)
	
	def visit_MenuDefinition(self, shape:static.MenuDefinition, node:InternalNode) -> LeafNode:
		ordinal = self.visit(shape.reader)
		try: within = shape.fields[ordinal]
		except KeyError: raise runtime.InvalidOrdinalError(shape.cursor_key, ordinal)
		else:
			try: branch = node.children[ordinal]
			except KeyError: branch = node.children[ordinal] = node_factory.visit(within)
			return self.visit(within, branch)
	
	def visit_SimpleReader(self, r:static.SimpleReader):
		try: return self.point[r.key]
		except KeyError:
			print("Reading", self.point)
			raise
	
	def visit_ComputedReader(self, r:static.ComputedReader):
		return self.env.read_magic(r.key, self.point)  # Method is up to the environment.
	
	def visit_DefaultReader(self, r:static.DefaultReader):
		return self.point.get(r.key, '_')  # Absent key becomes '_'; for cosmetic frames.

class NodeFilter(foundation.Visitor):
	""" Commonalities for finding matching nodes after-the-fact. """
	
	def visit_IsEqual(self, c: formulae.IsEqual, children: dict):
		if c.distinguished_value in children:
			yield c.distinguished_value, children[c.distinguished_value]
	
	def visit_IsInSet(self, c:formulae.IsInSet, children: dict):
		for k in c.including & children.keys():
			yield k, children[k]
	
	def visit_IsNotInSet(self, c:formulae.IsNotInSet, children: dict):
		for k in children.keys() - c.excluding:
			yield k, children[k]
	
	def visit_IsDefined(self, _:formulae.IsDefined, children: dict):
		return children.items()


class FindData(NodeFilter):
	""" Accumulate a list of matching (usually data) leaf indexes based on criteria. """
	
	def __init__(self, context: dict, selection: formulae.Selection):
		self.context = context
		self.criteria = selection.criteria
		self.found = []
	
	def visit_LeafDefinition(self, shape:static.LeafDefinition, node:LeafNode, remain:int):
		if remain == 0:
			self.found.append(node.begin)
	
	def visit_TreeDefinition(self, shape:static.TreeDefinition, node:InternalNode, remain:int):
		if self.common(shape.cursor_key, lambda o:shape.within, node, remain):
			for child in node.children.values():
				self.visit(shape.within, child, remain)

	def visit_FrameDefinition(self, shape:static.FrameDefinition, node:InternalNode, remain:int):
		key = shape.cursor_key
		if self.common(key, shape.fields.__getitem__, node, remain):
			# TODO: Maybe one day this will be a static analysis instead.
			raise runtime.AbsentKeyError(shape.cursor_key, self.context)
		
	def visit_MenuDefinition(self, shape:static.MenuDefinition, node:InternalNode, remain:int):
		if self.common(shape.cursor_key, shape.fields.__getitem__, node, remain):
			for ordinal, child in node.children.items():
				self.visit(shape.fields[ordinal], child, remain)

	def common(self, key:str, down:Callable[[str], static.ShapeDefinition], node:InternalNode, remain:int) -> bool:
		""" Commonalities among composite-type shapes; returns True if can't constrain. """
		if key in self.criteria:
			remain -= 1
			for ordinal, child in self.visit(self.criteria[key], node.children):
				self.visit(down(ordinal), child, remain)
		elif key in self.context:
			ordinal = self.context[key]
			child = node.children[ordinal]
			self.visit(down(ordinal), child, remain)
		else: return True
	
class LeafTour(foundation.Visitor):
	""" Walk a tree while keeping a cursor up to date; yield the leaf nodes. """
	
	def __init__(self, cursor:dict):
		self.cursor = cursor
	
	def visit_LeafDefinition(self, shape:static.LeafDefinition, node:LeafNode):
		yield node
	
	def visit_CompoundShapeDefinition(self, shape:static.CompoundShapeDefinition, node:InternalNode):
		for label, child_node in node.children.items():
			self.cursor[shape.cursor_key] = label
			yield from self.visit(shape.descend(label), child_node)
			del self.cursor[shape.cursor_key]
	
	def visit_Direction(self, direction:Direction):
		return self.visit(direction.shape, direction.tree)

class InternalTour(NodeFilter):
	"""
	Yield matching (internal, if possible) nodes.
	This is useful for merges, for outline specifications,
	and possibly for various other activities.
	"""
	
	def __init__(self, cursor: dict, selection: formulae.Selection):
		self.cursor = cursor
		self.criteria = selection.criteria
	
	def visit_LeafDefinition(self, shape:static.LeafDefinition, node:LeafNode, remain:int):
		if remain == 0: yield node
	
	def visit_CompoundShapeDefinition(self, shape:static.CompoundShapeDefinition, node:InternalNode, remain:int):
		if remain == 0: yield node
		else:
			if shape.cursor_key in self.criteria:
				predicate = self.criteria[shape.cursor_key]
				items = self.visit(predicate, node.children)
				remain -= 1
			else:
				items = node.children.items()
			for ordinal, child in items:
				self.cursor[shape.cursor_key] = ordinal
				yield from self.visit(shape.descend(ordinal), child, remain)
				del self.cursor[shape.cursor_key]
		
class Cartographer(foundation.Visitor):
	"""
	Contribute to the preparation of a properly-ordered list of leaf nodes.
	Seems to also be responsible for determining formats and formulas,
	collaborating with PlanState
	"""
	def __init__(self, begin:int, skin:veneer.PartialClassifier, patch:veneer.PartialClassifier):
		self.index = begin
		self.skin = skin
		self.patch = patch
	
	def enter_node(self, node:Node, state:veneer.PlanState):
		node.begin = self.index
		node.style_class = self.skin.classify(state)
		node.formula_class = self.patch.classify(state)
	
	def leave_node(self, node:InternalNode):
		node.size = self.index - node.begin
	
	def visit_LeafDefinition(self, shape: static.LeafDefinition, node: LeafNode, state:veneer.PlanState):
		self.enter_node(node, state)
		self.index += 1

	def _compound(self, shape:static.CompoundShapeDefinition, node:InternalNode, state:veneer.PlanState, schedule):
		def enter(label, is_first, is_last):
			prime = state.prime(shape.cursor_key, label, is_first, is_last)
			self.visit(shape.descend(label), node.children[label], prime)
		
		# Begin:
		self.enter_node(node, state)
		if len(schedule) == 1:
			# The only element is also the first and last element.
			enter(schedule[0], True, True)
		elif schedule:
			# There's a first, a possibly-empty middle, and a last element.
			enter(schedule[0], True, False)
			for i in range(1, len(schedule) - 1): enter(schedule[i], False, False)
			enter(schedule[-1], False, True)
		self.leave_node(node)
	
	def visit_TreeDefinition(self, shape: static.TreeDefinition, node: InternalNode, state: veneer.PlanState):
		schedule = sorted(node.children.keys(), key=state.environment.collation(shape.cursor_key))
		return self._compound(shape, node, state, schedule)
	
	def visit_FrameDefinition(self, shape:static.FrameDefinition, node:InternalNode, state:veneer.PlanState):
		schedule = shape.sequence
		return self._compound(shape, node, state, schedule)
		
	def visit_MenuDefinition(self, shape:static.MenuDefinition, node:InternalNode, state:veneer.PlanState):
		schedule = [k for k in shape.fields if k in node.children]
		return self._compound(shape, node, state, schedule)
		

class FreshNodeFactory(foundation.Visitor):
	"""
	Return a fresh Node subclass object according to whatever
	sort of shape definition we hand it.
	
	This is both visitor and singleton, because it has no particular state.
	However, it's simpler to just reuse the visitor class rather than
	build an overt `MonkeyPatch` class...
	"""
	
	def visit_LeafDefinition(self, shape:static.LeafDefinition):
		return LeafNode(shape.margin)
	
	def visit_CompoundShapeDefinition(self, shape:static.CompoundShapeDefinition):
		return InternalNode(shape.margin)
	
	def visit_FrameDefinition(self, shape:static.FrameDefinition):
		node = InternalNode(shape.margin)
		for label, child in shape.fields.items():
			node.children[label] = self.visit(child)
		return node

node_factory = FreshNodeFactory()



"""
Sequencing control for the cubicle compiler.
Basically, it's this module's job to coordinate the activities of
the front and middle ends to produce a loadable static object
which can later be fed to the dynamic module.
"""

from . import frontend, middle, static

def compile_string(string, *, filename=None) -> static.CubModule:
	parser = frontend.CoreDriver()
	declarations = parser.parse(string, filename=filename)
	if parser.errors:
		raise ValueError("Unable to compile cubicle module.") from None
	else:
		return middle.Transducer(parser.source).interpret(declarations)

def compile_path(path) -> static.CubModule:
	with open(path) as fh: string = fh.read()
	return compile_string(string, filename=path)

def main():
	import sys, argparse
	print("Not ready for prime time. Check back each Saturday night.", file=sys.stderr)
	exit(1)

if __name__ == "__main__":
	main()



"""
I face an odd pressure at the moment:

I'd eventually like to support a proper (if simple) expression language for
both label-type templates and formulas. Whatever had been in before was -- wrong.
So here's a static structure JUST FOR THAT MINI-LANGUAGE, independent of the
surrounding language of layout structures.

The structures in here are assumed to be final-form, pre-checked for any
necessary semantic constraints. Also, they are entirely passive data, like
instructions for some other machine to process. The strategy for doing so is
naturally to subclass `Visitor`, given in the boozetools `foundation` module.
"""

from typing import NamedTuple, List, Union, Dict, Container

class TextElement:
	""" ABC for things that can appear in EITHER a label OR a formula """

class LiteralText(TextElement, NamedTuple):
	text:str

class PlainOrdinal(TextElement, NamedTuple):
	""" The environment provides for how to make a value "plain" (human-friendly) given both axis-name and value. """
	axis: str

class RawOrdinal(TextElement, NamedTuple):
	""" In other words, the internal form of whatever ordinal. The environment need not be consulted. """
	axis: str

class Attribute(TextElement, NamedTuple):
	""" These have got to be computed. The environment must provides the method. """
	axis: str
	method: str



class Predicate:
	""" ABC for things we know how to test... """

class IsFirst(Predicate): pass
class IsLast(Predicate): pass
class IsEqual(Predicate, NamedTuple):
	distinguished_value: object
class IsInSet(Predicate, NamedTuple):
	including: frozenset
class IsNotInSet(Predicate, NamedTuple):
	excluding: frozenset
class IsDefined(Predicate): pass
class ComputedPredicate(Predicate, NamedTuple):
	cookie:str


class Selection(NamedTuple):
	criteria: Dict[str, Union[Predicate, List[Predicate]]]
	
	def projection(self, space:Container) -> "Selection":
		return Selection({k:p for (k, p) in self.criteria.items() if k in space})


class Boilerplate:
	""" ABC for things the language can specify for a cell. """

class BlankCell(Boilerplate): pass
THE_NOTHING = BlankCell()

class RawCell(Boilerplate, NamedTuple):
	axis:str

class Label(Boilerplate, NamedTuple):
	bits: List[TextElement]

class Formula(Boilerplate, NamedTuple):
	bits: List[Union[TextElement, Selection]]




"""
This file describes all the formatting instructions that come out of a definition in the
DSL, and how they are reconciled against each other.

Small cooperating parts are the order of the day...

"""

import operator
from typing import List, Container, Generic, TypeVar
from boozetools.support import foundation
from . import formulae, runtime

class PlanState(foundation.Visitor):
	def __init__(self, cursor:dict, first:frozenset, last:frozenset, environment:runtime.Environment):
		self.cursor = cursor
		self.first = first
		self.last = last
		self.environment = environment
		
	def prime(self, key, label, is_first, is_last):
		def tweak(some_set, is_member): return some_set | {key} if is_member else frozenset()
		return PlanState({**self.cursor, key:label}, tweak(self.first, is_first), tweak(self.last, is_last), self.environment)
	
	def visit_Selection(self, sel:formulae.Selection) -> bool:
		return all(self.visit(p, k) for k,p in sel.criteria.items())
	
	def visit_IsFirst(self, _, k): return k in self.first
	def visit_IsLast(self, _, k): return k in self.last
	def visit_IsEqual(self, p, k):
		try: return self.cursor[k] == p.distinguished_value
		except KeyError: return False
	def visit_IsInSet(self, p, k):
		try: return self.cursor[k] in p.including
		except KeyError: return False
	def visit_IsNotInSet(self, p, k):
		try: return not self.cursor[k] in p.excluding
		except KeyError: return False
	def visit_IsDefined(self, _, k):
		return k in self.cursor
	def visit_ComputedPredicate(self, p:formulae.ComputedPredicate, k:str):
		try: ordinal = self.cursor[k]
		except KeyError: return False
		else: return self.environment.test_predicate(p.cookie, ordinal)

T = TypeVar("T")
class Rule(Generic[T]):
	"""
	Associates zero or more selection criteria with {styling rules, formulae, etc.}.
	"""
	def __init__(self, selection: formulae.Selection, payload: T):
		self.selection = selection
		self.payload = payload


class PartialClassifier:
	"""
	In concept, a rule may be guarded by both horizontal and vertical predicates.
	Given a set of rules, this object builds a filter that is sensitive in one direction
	and permissive in the other direction. Whenever both orthogonal directions permit
	the same rule, that rule applies to corresponding cells in a notional grid.
	So the way this works is to number the DISTINCT sets of selected rules in either
	direction, and then the cross product of these distinct sets should be much easier
	to work with than recomputing a format (or hint) for each cell.
	"""
	def __init__(self, space:Container, rules:List[Rule]):
		self._relevant_predicates = [rule.selection.projection(space) for rule in rules]
		self._ec = foundation.EquivalenceClassifier()

	def classify(self, visitor:PlanState):
		return self._ec.classify(tuple(visitor.visit(ps) for ps in self._relevant_predicates))

	def mask(self, cls:int) -> frozenset:
		return self._ec.exemplars[cls]

class CrossClassifier:
	"""
	This makes the styling and hint selection algorithms more clear.
	"""
	def __init__(self, rules:List[Rule], hspace:Container, vspace:Container):
		self.across = PartialClassifier(hspace, rules)
		self.down = PartialClassifier(vspace, rules)
	def select(self, col_cls, row_cls):
		mask = map(operator.and_, self.across.mask(col_cls), self.down.mask(row_cls))
		return [i for i,b in enumerate(mask) if b]





"""
Packaging script for PyPI.
"""

import setuptools

setuptools.setup(
	name='cubicle',
	author='Ian Kjos',
	author_email='kjosib@gmail.com',
	version='0.8.4', # Might break a few things, probably won't.
	packages=['cubicle', ],
	package_dir = {'': 'src'},
	package_data={
		'cubicle': ['core.md',],
	},
	license='MIT',
	description='a high-level declarative domain-specific language for high-functioning, professional-looking, business-oriented numerical and graphical reporting',
	long_description=open('README.md').read(),
	long_description_content_type="text/markdown",
	url="https://github.com/kjosib/glowing-chainsaw",
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Development Status :: 3 - Alpha",
		"Intended Audience :: Developers",
		"Topic :: Office/Business",
		"Topic :: Software Development :: Interpreters",
		"Topic :: Software Development :: Libraries",
		
    ],
	python_requires='>=3.7',
	install_requires=[
		'xlsxwriter', 'booze-tools>=0.4.3'
	]
)


"""
Some utility functions and classes that should make life easier everywhere else.
"""
import pathlib, tempfile, pickle
from typing import Iterable
from xlsxwriter.utility import xl_rowcol_to_cell, xl_range


def make_range(col_run, row_run):
	if isinstance(col_run, int) and isinstance(row_run, int): return xl_rowcol_to_cell(row_run, col_run)
	else:
		if isinstance(col_run, int): left = right = col_run
		else: left, right = col_run
		if isinstance(row_run, int): top = bottom = row_run
		else: top, bottom = row_run
		return xl_range(top, left, bottom, right)


def collapse_runs(entries: Iterable[int]):
	""" Performs a simple run-length encoding. Runs of consecutive integers become <first,last> tuples. """
	
	def stash(): result.append(begin if begin == current else (begin, current))
	
	result = []
	traversal = iter(entries)
	try: begin = current = next(traversal)
	except StopIteration: return result
	for k in traversal:
		if k == current + 1: current = k
		else:
			stash()
			begin = current = k
	stash()
	return result

def tables(basis, doc) -> dict:
	"""
	Perhaps this routine belies a deficiency in the stack, but the object is to be able to
	compile the grammar only once (or whenever it changes) and use it over and over.
	In a fully-packaged solution a pre-pickled table may seem desirable, but for now
	this adaptive approach is better for hacking on.
	"""
	grammar_path = pathlib.Path(basis).parent/doc
	cache_path = pathlib.Path(tempfile.gettempdir())/(doc+'.pickle')
	def rebuild():
		from boozetools.macroparse import compiler
		result = compiler.compile_file(grammar_path, method='LR1')
		pickle.dump((grammar_stat.st_mtime, grammar_stat.st_size, result), open(cache_path, 'wb'))
		return result
	
	grammar_stat = grammar_path.stat()
	try: cache_stat = cache_path.stat()
	except FileNotFoundError: return rebuild()
	if grammar_stat.st_mtime > cache_stat.st_mtime: return rebuild()
	try: saved_mtime, saved_size, saved_table = pickle.load(open(cache_path, 'rb'))
	except: return rebuild()
	if (saved_mtime, saved_size) == (grammar_stat.st_mtime, grammar_stat.st_size): return saved_table
	else: return rebuild()
	
